/*
	*** USB or RS232 to CENTRONICS ***
	http://clx.freeshell.org/centronics.html
	This project is under Creative Common
	2015 CC BY-NC (Attribution-NonCommercial) 
	http://creativecommons.org/licenses/by-nc/4.0/
*/

#include <18f4550.h>

#fuses HSPLL,NOWDT,NOPROTECT,NOLVP,NODEBUG,USBDIV,PLL5,CPUDIV1,VREGEN
#use delay(clock=48MHz)
#use rs232(baud=9600, xmit=PIN_C6, rcv=PIN_C7)

#include "usb_desc_cdc.h"
#include <usb_cdc.h>
#include "defines.h"

// Outputs (pic to printer)
#bit _SELECT    = PORTA.0 // Caused to hang a Epson LX-400, so it's keep in high-Z
#bit _RESET     = PORTA.1 // cycle power to reset printer
#bit _AUTOFEED  = PORTA.5
#bit _IBM_INIT  = PORTB.5 // Sometimes ground, sometimes something else. We don't use it.
#byte DATAOUT   = PORTD   // Data pins, your text travels thru this.
#bit _STROBE    = PORTC.2 // Active low pulse used to transfer data. Must be between 0.5 and 500 microseconds for most printers.

// Inputs (printer to pic)
#bit _ACK       = PORTB.0 // Active low pulse indicates that data has been received and the printer is ready to accept more.
#bit  BUSY      = PORTB.1 // High indicates that the printer cannot receive data
#bit  PAPER_OUT = PORTB.2 // A high signal indicates that the printer is out of paper (Paper End)
#bit _FAULT     = PORTB.4 // Low indicates that the printer is in an off-line or error state
#bit  SELECT    = PORTB.3 // Indicate something from the printer

// Serial port related
#bit  TX        = PORTC.6 // output to RS232 port
#bit  RX        = PORTC.7 // input from RS232 port
#bit  CTS       = PORTE.1 // output to RS232 port
#bit  RTS       = PORTE.2 // input from RS232 port

// Miscellaneous stuff
#bit  PGC       = PORTB.6 // reserved for debugging
#bit  PGD       = PORTB.7 // reserved for debugging
#bit _MCLR      = PORTE.3 // reset input .. must be pulled up to VDD!
#bit  LED       = PORTC.1 // a cheap and usefull state indicator!

void init(void) {
	_AUTOFEED = 1;
	_STROBE = 1;

	TRISA = 0b00000001;
	TRISB = 0b00111111;
	TRISC = 0b10000000;
	TRISD = 0b00000000;
	TRISE = 0b00001100;
	setup_adc_ports(NO_ANALOGS);
	printf("\r\n# USB and RS232 to Centronics converter\r\n");
	printf("# http://clx.freeshell.org/centronics.html\r\n");
	printf("# Built: %s %s with ", __DATE__, __TIME__);
#ifdef __PCH__
	printf("PCH %s\r\n", __PCH__);
#else
	printf("PCM %s\r\n", __PCM__);
#endif

#if !(__USB_PIC_PERIF__)
	printf("# USBN: 0x%X\r\n", usbn_get_version());
#endif
	_RBPU = 0; // turn on weak pull-up on port B
	LED = 1;
	_RESET = 0;
	delay_ms(500);
	_RESET = 1;
	
}

void init_interrups(void){
	setup_timer_1(T1_INTERNAL|T1_DIV_BY_8); 
	enable_interrupts(INT_TIMER1);
	enable_interrupts(GLOBAL);
	enable_interrupts(INT_RDA);
}

#define BUFFER_SIZE 64
#define BUFFER_XON  55
#define BUFFER_XOFF 40

#define XON   0x11
#define XOFF  0x13
#define CTS_ENABLE 0
#define CTS_DISABLE 1
#define bkbhit (next_in!=next_out)

BYTE buffer[BUFFER_SIZE];
BYTE next_in = 0;
BYTE next_out = 0;
unsigned int8 buffer_freebytes = BUFFER_SIZE-1;
int1 buffer_xon_state = 1;

#int_rda
void serial_isr() {
	int t;
	buffer[next_in]=getc();
	buffer_freebytes--;
	t=next_in;
	next_in=(next_in+1) % BUFFER_SIZE;
	if(next_in==next_out){ // buffer full !
		next_in=t;
		buffer_freebytes = 0;
		return;
	}
	if (buffer_freebytes<BUFFER_XOFF){ 
		CTS = CTS_DISABLE;
		if (buffer_xon_state) {
			putc(XOFF);
			buffer_xon_state = 0;
		}
	}
}

BYTE bgetc(void) {
	BYTE c;
	if (!bkbhit){ return 0; }
	c=buffer[next_out];
	buffer_freebytes++;
	next_out=(next_out+1) % BUFFER_SIZE;
	if (buffer_freebytes>BUFFER_XON) {
		CTS = CTS_ENABLE;
		if (!buffer_xon_state){
			putc(XON);
			buffer_xon_state = 1;
		}
	}
	return(c);
}

int8 cnt_busy=30;

typedef enum {
	LP_DEFAULT = 0,
	LP_IDLE,
	LP_PRINTER_NOT_OK,
	LP_ENUMERATED,
	LP_CDC,
	LP_WAIT,
	LP_RS232
} t_ledpattern_state;
t_ledpattern_state ledpattern_state=LP_DEFAULT;

int clk1_div=0;
#int_timer1
void clk1(void){
	static int8 led_pattern;
	static t_ledpattern_state last_ledpattern_state=LP_DEFAULT;
	if (cnt_busy) { cnt_busy--; }
	
	if (clk1_div--){ return; }
	clk1_div = 3;
	if (ledpattern_state != last_ledpattern_state) {
		switch (ledpattern_state){
			default: 
				led_pattern = 0b10100111; break;
			case LP_IDLE:
				led_pattern = 0b00000001; break;
			case LP_PRINTER_NOT_OK:
				led_pattern = 0b10101000; break;
			case LP_ENUMERATED:
				led_pattern = 0b10111000; break;
			case LP_CDC:
				led_pattern = 0b11111111; break;
			case LP_WAIT:
			case LP_RS232:
				led_pattern = 0b00000000; break;
		}
		last_ledpattern_state = ledpattern_state;
	}
	rotate_left(&led_pattern, 1);
	LED = led_pattern&1;
}

// when called periodically, displays debugging information 
// about enumeration and connection states.  
void usb_debug_task(void) {
	static int8 last_connected;
	static int8 last_enumerated;
	static int1 last_dte_present, last_active;
	int8 connected;
	int8 enumerated;
	static int8 last_cdc;
	int8 cdc;

	connected=usb_attached();
	enumerated=usb_enumerated();
	cdc=usb_cdc_connected();

	if (usb_cdc_carrier.dte_present) {
		if (!last_dte_present) { 
			printf("[DTE IS NOW PRESENT]\r\n");
			last_dte_present=1;
			ledpattern_state = LP_CDC;
		}
	}
	else {
		if (last_dte_present) { 
			printf("[DTE IS NOW ABSENT]\r\n");
			last_dte_present=0;
			ledpattern_state = LP_IDLE;
		}
	}
	if (usb_cdc_carrier.active) {
		if (!last_active) {
			printf("[DTE ACTIVATED]\r\n");
			last_active=1;
			ledpattern_state = LP_CDC;
		}
	}
	else {
		if (last_active) { 
			printf("[DTE DESACTIVATED]\r\n"); 
			last_active=0; 
			ledpattern_state = LP_IDLE;
		}
	}
	
	if (connected && !last_connected) {
		printf("[USB connected, waiting for enumaration...]\r\n");
		ledpattern_state = LP_WAIT;
	}
	if (!connected && last_connected){
		printf("[USB disconnected, waiting for connection...]\r\n");
		ledpattern_state = LP_WAIT;
	}
	if (enumerated && !last_enumerated){
		printf("[USB enumerated by PC/HOST]\r\n");
		ledpattern_state = LP_ENUMERATED;
	}
	if (!enumerated && last_enumerated){
		printf("[USB unenumerated by PC/HOST, waiting for enumeration...]\r\n");
		ledpattern_state = LP_WAIT;
	}
	if (cdc && !last_cdc) {
		printf("[Serial program initiated on USB<->UART COM Port]\r\n");
		ledpattern_state = LP_IDLE;
	}

	last_connected=connected;
	last_enumerated=enumerated;
	last_cdc=cdc;
}

void serial_sanity(void) {
		if (OERR) {
			// in case of buffer overflow, we restart the link.
			CREN = 0;
			CREN = 1;
			if (buffer_xon_state){
				putc(XON);
			}
		}

		if (FERR) {
		}
}

void lpt_putc(char c) {
	DATAOUT = c;
	_STROBE = 0;
	delay_us(250);
	_STROBE = 1;
}

void report_printer_availability(void){
	static int1 last_rx_ok = 0;
	static int1 last_paper_out=0, last_fault=0, last_busy=0;
	int1 rx_ok = 1;
	
	if (BUSY){ 
		if (!cnt_busy) { 
			rx_ok = 0;
			if (!last_busy) { 
				printf("[BUSY FOR LONG TIME]\r\n");
				last_busy = 1;
			}
		}
	} 
	else { 
		if (last_busy) { 
			printf("[NOT BUSY ANYMORE]\r\n");
			last_busy = 0; 
		}
		cnt_busy = 20;
	}
	
	if(!_FAULT) {
		rx_ok = 0;
		if (!last_fault){ last_fault = 1; printf("[FAULT]\r\n"); }
	}
	else {
		if (last_fault) { last_fault = 0; printf("[NOT FAULT]\r\n"); }
	}
	if (PAPER_OUT){
		rx_ok = 0;
		if (!last_paper_out) { last_paper_out = 1; printf("[PAPER OUT]\r\n"); }
	}
	else {
		if (last_paper_out) { last_paper_out = 0; printf("[PAPER PRESENT]\r\n"); }
	}
	
	if (buffer_xon_state){
		if (rx_ok && !last_rx_ok) { 
			putc(XON); 
			ledpattern_state = LP_IDLE;
		}
		if (!rx_ok && last_rx_ok) {
			putc(XOFF);
		}
		CTS = rx_ok?CTS_ENABLE:CTS_DISABLE;
		last_rx_ok = rx_ok;
	}
	
	if (!rx_ok){
		ledpattern_state = LP_PRINTER_NOT_OK;
	}
}

/*
void lpt_test(char c) {
	do {
		report_printer_availability();
	} while(BUSY);
	lpt_putc(c);
}
*/

void main(void) {
	char c;

	init();
	init_interrups();

	//lpt_test("The quick brown fox jumps over the lazy dog\r\n");

	usb_init_cs();
	
	for(;;) { // main loop
		usb_task();
		usb_debug_task();

		// print some verbose informations about the printer's states on the serial port only
		report_printer_availability();
		
		if (!BUSY && !PAPER_OUT && _FAULT) { // ready to print
			if (bkbhit) { // character from RS232				
				LED=1; ledpattern_state = LP_RS232;
				c=bgetc();
				lpt_putc(c);
				if (usb_cdc_carrier.dte_present) { usb_cdc_putc(c); }
			}
			if (usb_cdc_kbhit()) { // character from USB CDC
				LED=0; ledpattern_state = LP_CDC;
				c=usb_cdc_getc();
				lpt_putc(c);
				putc(c);
			}
		}
		serial_sanity();
	} // end main loop
}

/*
>>> Warning 216 "main.c" Line 238(22,23): Interrupts disabled during call to prevent re-entrancy:  (usb_token_reset)
>>> Warning 216 "main.c" Line 238(22,23): Interrupts disabled during call to prevent re-entrancy:  (usb_tbe)
>>> Warning 216 "main.c" Line 238(22,23): Interrupts disabled during call to prevent re-entrancy:  (usb_cdc_flush_out_buffer)
>>> Warning 216 "main.c" Line 238(22,23): Interrupts disabled during call to prevent re-entrancy:  (usb_cdc_get_discard)
*/
#ignore_warnings 216